#-*- coding: utf-8 -*-
#@authors: Pau Ferrer Ocaña
#@authors: Modified TICxCAT
#This file is part of Trivial Freezer.

#Trivial Freezer is an easy freezer for user profiles and desktop in linux.
#Copyright (C) 2009  Pau Ferrer Ocaña

#Trivial Freezer is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#Trivial Freezer is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with Trivial Freezer.  If not, see <http://www.gnu.org/licenses/>.

from TFglobals import *

import os
import tarfile
import re
import shutil
import paramiko, pwd
import subprocess
import tempfile

_ = load_locale()


def move(src,dst,path): #path argument added by TICxCAT
    "Moves a file without overwritting it, appends a number at the end of the name"

    auxPath = 0

    dstComplete = dst
    dst = dst	+ '/' + path #Added by TICxCAT

    #If the destination already exists, change the filename appending a number
    while os.path.exists(dst):
        dstComplete = dst + "_" + str(auxPath)
        auxPath = auxPath + 1
	debug(dst,DEBUG_HIGH) #Added by TICxCAT
	dst=dstComplete #Added by TICxCAT to move out of while

    shutil.move(src, dstComplete)

class user_frozen:
    "Class to create and restore the user frozen profiles"

    #Profile Name
    name = ""
    #Profile Filters
    filters = []
    #If specified, what file use to save/restore
    source = ""
    #If specified, Lost+found directory, if not, use defaults
    deposit = DEFAULT_DEPOSIT
    #Command to execute after restoring
    execute = ""

    #User name
    username = ""
    #User Home Directory
    homedir = ""
    #User ID
    uid = 0
    #Group ID
    gid = 0

    #If specified, Hostname/IP of the ssh/nfs server where the user will be restored
    hostname = ""
    #If specified, Port of the ssh/nfs server where the user will be restored
    port = "22"
    #If specified, user of the ssh/nfs server who has privileges to execute tfreezer
    server_user = "tfreezer"

    def __init__(self, title, deposit, rules, source = "", execute = ""):
        "Initializes a user_frozen with the selected options"
        self.filters = [] #Line added by TICxCAT to re-initialize filters for each user
        self.name =  title
        if deposit == '':
            deposit = DEFAULT_DEPOSIT
        self.deposit = deposit
        for rule in rules:
            self.filters.append([rule.action,rule.filter])

    def create_tar(self):
        "Creates a tar for the frozen user"
        # If the source is specified, no need to create nothing
        if len(self.source) > 0: return

        debug("Creating backup for " + self.username + " with profile " + self.name,DEBUG_LOW)

        #Can't access home directory
        if not os.access(self.homedir, os.R_OK):
            print_error("on create_tar. Can't access " + self.homedir)
            raise

        # Start preparing options for the command we are going to run
        backup_path = os.path.join(TAR_DIRECTORY, TAR_HOMES)
        rsync_command = ['rsync','-a','--exclude='+os.path.basename(self.deposit)]
        for f in [b for b in self.filters if b[0] == ACTION_LOST]:
            rsync_command.append('--exclude='+str(f[1]))
        # we first load the includes
        for f in [b for b in self.filters if b[0] == ACTION_RESTORE]:
            rsync_command.append('--include='+f[1])
        # then the excludes
        for f in [b for b in self.filters if b[0] == ACTION_KEEP or b[0] == ACTION_ERASE]:
            rsync_command.append('--exclude='+str(f[1]))
        # if self.hostname is not set, then we omit the --address option
        if self.hostname == '':
            rsync_command.extend([self.homedir.rstrip('/')+'/',os.path.join(backup_path,self.username)])
        else:
            rsync_command.extend(['--address',self.hostname,self.homedir,backup_path])
        # if more than 1 filter is set, then is not a total freeze
        if len(self.filters) > 1:
            #print rsync_command
            process = subprocess.Popen(rsync_command)
            process.wait()
        else:
            # execute the rsync. --delete will delete every non-existing file/folder in the
            # backupdir.
            process = subprocess.Popen(['rsync',
                        '-a',
                        '--delete',
                        '--address',
                        self.hostname,
                        self.homedir,
                        backup_path])
            process.wait()


    def restore_tar(self):
        "Restores a tar for the frozen user. If the hostname is specified, it \
        restores on an external server"
        #If the hostname is specified, restore in the external server
        if len(self.hostname) > 0:
            self.__restore_external_tar()
            return

        debug("Restoring backup for " + self.username + " with profile " + self.name,DEBUG_LOW)
        # If there are specific filters rsync without --delete

        backup_path = os.path.join(TAR_DIRECTORY, TAR_HOMES)
        rsync_command = ['rsync','-a','--exclude='+os.path.basename(self.deposit)]
        for f in [b for b in self.filters if b[0] == ACTION_LOST]:
            rsync_command.append('--exclude='+str(f[1]))
        # we first load the includes
        for f in [b for b in self.filters if b[0] == ACTION_RESTORE]:
            rsync_command.append('--include='+f[1])
        # then the excludes
        for f in [b for b in self.filters if b[0] == ACTION_KEEP or b[0] == ACTION_ERASE]:
            rsync_command.append('--exclude='+str(f[1]))
        # if self.hostname is not set, then we omit the --address option
        if self.hostname == '':
            rsync_command.extend(['--delete',os.path.join(backup_path,self.username).rstrip('/')+'/',self.homedir.rstrip('/')])
        else:
            rsync_command.extend(['--delete','--address',self.hostname,backup_path+'/',self.homedir])
        # if more than 1 filter is set, then is not a total freeze
        if len(self.filters) > 1:
            dir = os.path.join (TAR_DIRECTORY, TAR_REPOSITORY)
            tarpath = os.path.join (dir, self.source)

            #If the deposit is in the home directory, replace with the correct home
            self.deposit = str(self.deposit.replace('~',self.homedir,1))
            if self.deposit.endswith("/"):
                self.deposit = self.deposit[:-1]

            #Create the deposit directory, if exists, doesn't matter...
            try:
                os.makedirs(self.deposit,0755)
            except OSError , (errno, strerror):
                pass
            else:
                #If the deposit is inside a home directory, change the owner to the user
                if self.deposit.startswith(self.homedir):
                    os.chown(self.deposit, self.uid, self.gid)

            #Apply the MAINTAIN and LOST+FOUND filters
            #self.__apply_filters(self.homedir)
            # rsync_movelost = ['rsync','-a','--exclude='+os.path.basename(self.deposit)]
            # for f in [b for b in self.filters if b[0] == ACTION_LOST]:
            #     rsync_movelost.append('--include='+str(f[1]))
            # rsync_movelost.extend(['--exclude=*','--remove-source-files',self.homedir+'/',self.deposit])
            #print self.filters
            for f in [b for b in self.filters if b[0] == ACTION_LOST]:
                mv_command = ['mv '+str(f[1])+' '+self.deposit+'/']
                #print mv_command 
                try:
                    process = subprocess.Popen(mv_command,shell=True,cwd=self.homedir)
                    process.wait()
                except:
                    pass

            # Apply the delete filters
            for f in [b for b in self.filters if b[0] == ACTION_ERASE]:
                rm_command = ['rm -rf ' + str(f[1])]
                try:
                    process = subprocess.Popen(rm_command,shell=True,cwd=self.homedir)
                    process.wait()
                except:
                    pass


            #process = subprocess.Popen(rsync_movelost)
            #print rsync_movelost
            if len([b for b in self.filters if b[0] == ACTION_KEEP or b[0] == ACTION_ERASE or b[0] == ACTION_RESTORE]) >= 1:
                process = subprocess.Popen(rsync_command)
                process.wait()



#        if len(self.filters) > 1:
#            #Apply the MAINTAIN and LOST+FOUND filters
#            self.__apply_filters(self.homedir)
#            backup_path = os.path.join(TAR_DIRECTORY, TAR_HOMES, self.username)
#            process = subprocess.Popen(['rsync', '-a', '--delete', '--address',
#                self.hostname, backup_path + '/', self.homedir])
#            process.wait()
        # if there aren't any filter, we supose it's a total recovery and
        # rsync is called with the --delete parameter. All files that are not
        # in the backup will be deleted.
        else:
            backup_path = os.path.join(TAR_DIRECTORY, TAR_HOMES, self.username)
            process = subprocess.Popen(['rsync', '-a', '--delete', '--address',
                self.hostname, backup_path + '/', self.homedir])
            process.wait()
        # #If the source is not specified, use the user tar
        # if len(self.source) < 1:
        #     dir = os.path.join (TAR_DIRECTORY, TAR_HOMES)
        #     tarpath = os.path.join (dir, self.username + TAR_EXTENSION)
        # #If the source is specified, use it
        # else:
        #     dir = os.path.join (TAR_DIRECTORY, TAR_REPOSITORY)
        #     tarpath = os.path.join (dir, self.source)

        #     #If the deposit is in the home directory, replace with the correct home
        #     self.deposit = str(self.deposit.replace('~',self.homedir,1))
        #     if self.deposit.endswith("/"):
        #         self.deposit = self.deposit[:-1]

        #     #Create the deposit directory, if exists, doesn't matter...
        #     try:
        #         os.makedirs(self.deposit,0755)
        #     except OSError , (errno, strerror):
        #         pass
        #     else:
        #         #If the deposit is inside a home directory, change the owner to the user
        #         if self.deposit.startswith(self.homedir):
        #             os.chown(self.deposit, self.uid, self.gid)

        #     #Apply the MAINTAIN and LOST+FOUND filters
        #     self.__apply_filters(self.homedir)

        # #Execute the command
        debug ('EXECUTING: ' + self.execute,DEBUG_HIGH)
        result = os.popen(self.execute).read()
        for line in result.splitlines():
            debug ('RESULT: ' + line , DEBUG_HIGH)

    def __restore_external_tar(self):
        "Restores a tar on an external server"

        #If the hostname is localhost, I'm the server, so all is locally
        #This kind of restoring is not permitted to avoid loops
        if self.hostname == 'localhost':
            print_error(_("Localhost external restoration not permitted to avoid loops"),WARNING)
            return

        debug("Restoring external backup for " + self.username + " with profile " + self.name + " on "+self.hostname,DEBUG_LOW)

        #Get the root hoem directory
        roothome = pwd.getpwuid(0).pw_dir
        try:
            #Read the private key SSH file and the SSH Client
            pkey = paramiko.DSSKey.from_private_key_file(roothome + '/'+ID_DSA_PATH,"")
            ssh = paramiko.SSHClient()
            try:
                #Load the known hosts for the root user
                ssh.load_system_host_keys(roothome + '/'+KNOWN_HOSTS_PATH)
            except:
                pass
            #Connect to the server with the root user. Do not look for keys, use the specified pkey
            ssh.connect(self.hostname,int(self.port),username=self.server_user,pkey=pkey,look_for_keys=False)
        except Exception , e:
            debug("Exception " + str(type(e)) + ": " + str(e), DEBUG_LOW)
            print_error(_("Can't connect to the server, please review your settings"))
            return

        #Create the command line
        if NEEDS_SUDO:
            #command = 'sudo tfreezer -a -r ' + self.username + ' -d ' + str(get_debug_level()) + ' 2>&1'
	    command = 'sudo tfreezer -r ' + self.username + ' -d ' + str(get_debug_level()) + ' 2>&1' #Modified by TICxCAT
        else:
            #command = 'tfreezer -a -r ' + self.username + ' -d ' + str(get_debug_level()) + ' 2>&1'
            command = 'tfreezer -r ' + self.username + ' -d ' + str(get_debug_level()) + ' 2>&1' #Modified by TICxCAT

        debug("Executing command " +command + " on server", DEBUG_LOW)

        #TOERASE 2
        import time
        start = time.time()

        #Run the command and print its results in debug mode.
        try:
            stdin,stdout,stderr = ssh.exec_command(command)
            for line in stdout.readlines():
                debug("Server: " + line,DEBUG_MEDIUM)
            stdout.close()
        except Exception , e:
            debug("Exception " + str(type(e)) + ": " + str(e), DEBUG_LOW)
            print_error(_("Can't execute the command"))

        #TOERASE 2
        end = time.time()
        print "Time elapsed restoring = ", end - start, "seconds"

        ssh.close()
        return


    def __exclude_from_tar(self, path):
        "Files to be excluded from the tar"
        "Applies the KEEP, ERASE, and RESTORE filters"
        "Returns True to exlude and False to include"

        if get_thread_killed():
            return False

        #Cut the path over the home directory
        path = path[len(self.homedir)+1:]

        #If the path and the filters are not empty, carry on
        if len(path) < 1 or len(self.filters) < 1:
            return False

        #Apply every filter
        for filter in self.filters:
            #if get_thread_killed():
            #    return False

            #If the path matches the filter
            if re.search(filter[1],path) != None:
                #Take the action of the filter
                action = filter[0]

                #For KEEP and ERASE action, exclude from the tar
                if action == ACTION_KEEP or action == ACTION_ERASE:
                    debug("Exclude  path: " + path,DEBUG_HIGH)
                    return True
                #For RESTORE action, include in the tar
                elif action == ACTION_RESTORE:
                    debug("Maintain path: " + path,DEBUG_HIGH)
                    return False
                #The LOST action does not affect

        #Default option is KEEP, so return True
        debug("Exclude  path: " + path,DEBUG_HIGH)
        return True


    def __restore_or_erase(self,path):
        "Returns True to carry on or erase and False to maintain"

        #Do nothing with the deposit if it's inside a home directory
        if path.startswith(self.deposit):
        	debug('Deposit path: ' + path,DEBUG_MEDIUM)
        	return False

        #Not erase the mount dirs different of homedir
        if os.path.ismount(path) and path != self.homedir:
        	debug('Mount point: ' + path,DEBUG_MEDIUM)
        	return False
        #Added by TICxCAT to maintain complete path for lost objects
        complete_path = path #TICxCAT modification
        #Cut the path over the home directory
        path = path[len(self.homedir)+1:]

        #Go on with the home directory (not to be erased, but go on)
        if len(path) <= 0:  return True

        #For every filter
        for filter in self.filters:
        	#If the path matches the filter
            if re.search(filter[1],re.escape(path)) != None:
        		#Take the action of the filter
        		action = filter[0]
        		#For KEEP action, do not remove it
        		if action == ACTION_KEEP:
        			debug('Keep path: ' + path,DEBUG_HIGH)
        			return False
        		#For LOST action, move it to deposit
        		elif action == ACTION_LOST:
        			debug('Lost path: ' + path,DEBUG_HIGH)
        			move(complete_path, self.deposit,path) #Modified by TICxCAT to maintain complete path for lost objects
        			return False
        		#For RESTORE and ERASE action, remove
        		elif action == ACTION_RESTORE or action == ACTION_ERASE:
        			return True

        #Default action is KEEP, so return false
        debug('Keep path (default): ' + path,DEBUG_HIGH)
        return False


    def __apply_filters(self, dirname):
        "Applies filters for restoring a tar"

        #It has to be erased? (Directory)
        if self.__restore_or_erase(dirname):
            #YES! Erase it, but first look inside it...
            files = os.listdir(dirname)
            for file in files:
                path = os.path.join (dirname, file)
                #If it's a directory or a link, recursive call
                if os.path.isdir(path) and not os.path.islink(path):
                    self.__apply_filters(path)
                #If it's a file, see if it has to be removed
                else:
                    debug('Removing file: ' + path,DEBUG_HIGH)
                    #It has to be erased? (File)
                    if self.__restore_or_erase(path):
                        os.unlink(path)

            #If we're not in the home directory and is an empty directory, erase it
            if dirname != self.homedir:
                if len(os.listdir(dirname)) == 0:
                    debug('Removing directory: ' + dirname,DEBUG_HIGH)
                    os.rmdir(dirname)
